module data_module

  use kinddefs,    only : dp, jp, dqp, odp
  use interp_defs, only : sendrecv_type

  implicit none

  private

  public :: load_data

  integer, public :: colored_sweeps, max_colored_sweeps, neq0, nia, nja, n_eqns
  integer, public :: dq_dim, nr, nm, n_sweeps, neqmax, nb, solve_backwards
  integer, public :: outer_sweeps

  integer, dimension(:),   allocatable, public :: iam, jam
  integer, dimension(:),   allocatable, public :: color_boundary_end
  integer, dimension(:,:), allocatable, public :: color_indices

  real(dp), public :: golden

  real(dp),  dimension(:,:),   allocatable, public :: res
  real(dqp), dimension(:,:),   allocatable, public :: dq
  real(jp),  dimension(:,:,:), allocatable, public :: a_diag_lu
  real(odp), dimension(:,:,:), allocatable, public :: a_off

  character(len=80), public :: datafile

  type(sendrecv_type), dimension(:), allocatable, public :: sr

  logical, parameter, public :: timing = .true.

contains

  subroutine load_data()

    use lmpi, only : lmpi_master, lmpi_id, array_of_statuses, lmpi_nproc,      &
                     lmpi_synchronize

    integer :: sr_size, dim1, dim2, dim3, dim4, dim5, dim6, dim7, dim8, i, j
    integer :: dim9, dim10, ipe

    if ( lmpi_master ) write(*,*) 'Loading data...'

    do ipe = 0, lmpi_nproc-1
      if ( ipe == lmpi_id ) then
        open(66,file=trim(datafile),form='unformatted',status='old',           &
             access='stream')

        read(66) colored_sweeps, max_colored_sweeps, neq0, nia, nja, n_eqns,   &
                 dq_dim, nr, nm, n_sweeps, neqmax, nb, solve_backwards,        &
                 outer_sweeps, sr_size

        write(*,'(i0,a,i0)') lmpi_id,                                          &
                             ' Number of block 5x5 equations in data file: ',  &
                             neq0

        allocate(color_indices(2,colored_sweeps))
        read(66) color_indices

        allocate(color_boundary_end(colored_sweeps))
        read(66) color_boundary_end

        allocate(iam(nia))
        read(66) iam

        allocate(jam(nja))
        read(66) jam

        allocate(res(nr,neq0))
        read(66) res

        allocate(dq(dq_dim,neqmax))
        read(66) dq

        allocate(a_diag_lu(nm,nm,neq0))
        read(66) a_diag_lu

        allocate(a_off(nm,nm,nja))
        read(66) a_off

        allocate(sr(sr_size))

        do i = 1, sr_size
          read(66) dim1, dim2, dim3, dim4, dim5, dim6, dim7, dim8, dim9, dim10
          allocate(sr(i)%sendproc(max(1,dim1)))
          allocate(sr(i)%recvproc(max(1,dim2)))
          allocate(sr(i)%sendindex(max(1,dim3)))
          allocate(sr(i)%recvindex(max(1,dim4)))
          allocate(sr(i)%double_mat_senddata(max(1,dim5),max(1,dim6)))
          allocate(sr(i)%double_mat_recvdata(max(1,dim7),max(1,dim8)))
          allocate(array_of_statuses(max(1,dim9),max(1,dim10)))
          do j = 1, dim1
            read(66) sr(i)%sendproc(j)
          end do
          do j = 1, dim2
            read(66) sr(i)%recvproc(j)
          end do
          do j = 1, dim3
            read(66) sr(i)%sendindex(j)
          end do
          do j = 1, dim4
            read(66) sr(i)%recvindex(j)
          end do
          sr(i)%double_mat_allocated = .true.
        end do

        read(66) golden

        close(66)
      endif
      call lmpi_synchronize
    end do

    if ( lmpi_master ) write(*,*) 'Done loading data...'

  end subroutine load_data

end module data_module


module solve_module

  implicit none
  private
  public :: point_solve_5

contains

  subroutine point_solve_5()

    use data_module, only : colored_sweeps, color_indices, max_colored_sweeps, &
                            neq0, nia, nja, iam, jam, n_eqns, solve_backwards, &
                            dq_dim, nr, nm, n_sweeps, neqmax, res, dq,         &
                            a_diag_lu, a_off, color_boundary_end, sr
    use kinddefs,    only : dqp, dp
    use lmpi_app,    only : lmpi_start_xfer, lmpi_complete_xfer

    integer :: j,n,sweep,icol,istart,iend,start,end,ipass
    integer :: color, loop_colored_sweeps
    integer :: sweep_start, sweep_end, sweep_stride

    real(dqp) :: f1,f2,f3,f4,f5

  continue

    loop_colored_sweeps = max_colored_sweeps

    sweep_start  = 1
    sweep_end    = loop_colored_sweeps
    sweep_stride = 1

    if ( solve_backwards > 1 .or. solve_backwards < -1 ) then
      sweep_start  = loop_colored_sweeps
      sweep_end    = 1
      sweep_stride = -1
    endif

    if ( n_eqns <= 0 ) then
      sweep_start  = 1
      sweep_end    = 2
      sweep_stride = -1
    endif

    !$omp parallel default(shared)

    sweeping : do sweep = 1, n_sweeps

      color_sweeps : do color = sweep_start, sweep_end, sweep_stride

        pass_loop : do ipass = 1, 2

          if ( color > colored_sweeps ) then
            start = 1
            end   = 0
          else
            select case(ipass)
            case(1)
              if ( color_boundary_end(color) == 0 ) then
                start = 1
                end   = 0
              else
                start = color_indices(1,color)
                end   = color_boundary_end(color)
              endif
            case(2)
              if ( color_boundary_end(color) == 0 ) then
                start = color_indices(1,color)
                end   = color_indices(2,color)
              else
                start = color_boundary_end(color)+1
                end   = color_indices(2,color)
              endif
            end select
          endif

          !$omp do private(f1,f2,f3,f4,f5,n,j,icol,istart,iend) schedule(auto)
          rhs_solve : do n = start, end

            if ( solve_backwards > 0 ) then
              f1 = -res(1,n)
              f2 = -res(2,n)
              f3 = -res(3,n)
              f4 = -res(4,n)
              f5 = -res(5,n)
            else
              f1 = res(1,n)
              f2 = res(2,n)
              f3 = res(3,n)
              f4 = res(4,n)
              f5 = res(5,n)
            end if

            istart = iam(n)
            iend   = iam(n+1)-1

            do j = istart,iend
              icol = jam(j)

              f1 = f1   - a_off(1,1,j)*dq(1,icol)
              f2 = f2   - a_off(2,1,j)*dq(1,icol)
              f3 = f3   - a_off(3,1,j)*dq(1,icol)
              f4 = f4   - a_off(4,1,j)*dq(1,icol)
              f5 = f5   - a_off(5,1,j)*dq(1,icol)

              f1 = f1   - a_off(1,2,j)*dq(2,icol)
              f2 = f2   - a_off(2,2,j)*dq(2,icol)
              f3 = f3   - a_off(3,2,j)*dq(2,icol)
              f4 = f4   - a_off(4,2,j)*dq(2,icol)
              f5 = f5   - a_off(5,2,j)*dq(2,icol)

              f1 = f1   - a_off(1,3,j)*dq(3,icol)
              f2 = f2   - a_off(2,3,j)*dq(3,icol)
              f3 = f3   - a_off(3,3,j)*dq(3,icol)
              f4 = f4   - a_off(4,3,j)*dq(3,icol)
              f5 = f5   - a_off(5,3,j)*dq(3,icol)

              f1 = f1   - a_off(1,4,j)*dq(4,icol)
              f2 = f2   - a_off(2,4,j)*dq(4,icol)
              f3 = f3   - a_off(3,4,j)*dq(4,icol)
              f4 = f4   - a_off(4,4,j)*dq(4,icol)
              f5 = f5   - a_off(5,4,j)*dq(4,icol)

              f1 = f1   - a_off(1,5,j)*dq(5,icol)
              f2 = f2   - a_off(2,5,j)*dq(5,icol)
              f3 = f3   - a_off(3,5,j)*dq(5,icol)
              f4 = f4   - a_off(4,5,j)*dq(5,icol)
              f5 = f5   - a_off(5,5,j)*dq(5,icol)

            end do

! Forward...sequential access to a_diag_lu.

            f2 = f2 - a_diag_lu(2,1,n)*f1
            f3 = f3 - a_diag_lu(3,1,n)*f1
            f4 = f4 - a_diag_lu(4,1,n)*f1
            f5 = f5 - a_diag_lu(5,1,n)*f1

            f3 = f3 - a_diag_lu(3,2,n)*f2
            f4 = f4 - a_diag_lu(4,2,n)*f2
            f5 = f5 - a_diag_lu(5,2,n)*f2

            f4 = f4 - a_diag_lu(4,3,n)*f3
            f5 = f5 - a_diag_lu(5,3,n)*f3

            f5 = f5 - a_diag_lu(5,4,n)*f4

! Backward...sequential access to a_diag_lu.

            dq(5,n) = f5 * a_diag_lu(5,5,n)
            f1 = f1 - a_diag_lu(1,5,n)*dq(5,n)
            f2 = f2 - a_diag_lu(2,5,n)*dq(5,n)
            f3 = f3 - a_diag_lu(3,5,n)*dq(5,n)
            f4 = f4 - a_diag_lu(4,5,n)*dq(5,n)

            dq(4,n) = f4 * a_diag_lu(4,4,n)
            f1 = f1 - a_diag_lu(1,4,n)*dq(4,n)
            f2 = f2 - a_diag_lu(2,4,n)*dq(4,n)
            f3 = f3 - a_diag_lu(3,4,n)*dq(4,n)

            dq(3,n) = f3 * a_diag_lu(3,3,n)
            f1 = f1 - a_diag_lu(1,3,n)*dq(3,n)
            f2 = f2 - a_diag_lu(2,3,n)*dq(3,n)

            dq(2,n) = f2 * a_diag_lu(2,2,n)
            f1 = f1 - a_diag_lu(1,2,n)*dq(2,n)

            dq(1,n) = f1 * a_diag_lu(1,1,n)

          end do rhs_solve
          !$omp end do

          select case(ipass)
          case(1)
            call lmpi_start_xfer(dq,sr_opt=sr(color))
          case(2)
            call lmpi_complete_xfer(dq,sr_opt=sr(color))
          end select

        end do pass_loop

      end do color_sweeps

    end do sweeping

    !$omp end parallel

  end subroutine point_solve_5


end module solve_module


program main

  use kinddefs,        only : dp
  use solve_module,    only : point_solve_5
  use ieee_arithmetic, only : ieee_is_nan
  use data_module,     only : load_data, golden, dq, neq0, datafile, timing,   &
                              dq_dim, nb, outer_sweeps
  use lmpi,            only : lmpi_start, lmpi_id, lmpi_finalize, lmpi_master, &
                              lmpi_bcast, lmpi_reduce, lmpi_synchronize

  implicit none

  integer :: i, j, neq

  !integer(kind=4) :: t1, t2, t3, t4
  double precision :: t1, t2, t3, t4

  real(dp) :: rms, rms_sum

  real(dp), parameter :: tol = 1.e-07_dp

  character(len=80) :: rank

  !DOUBLE PRECISION MPI_WTIME
  double precision omp_get_wtime

continue

  call lmpi_start

  write(rank,'(i0)') lmpi_id
  datafile = 'linear_system.dat.' // trim(rank)

! Load the dataset for testing

  call load_data()

  call lmpi_synchronize()

  !call system_clock(t1)
  !if (lmpi_master .and. timing) t1 = mpi_wtime()
  if (lmpi_master .and. timing) t1 = omp_get_wtime()

! Solve the mock problem

  if ( lmpi_master ) write(*,*) 'Solving Ax=b...'

  do i = 1, outer_sweeps
    !if ( lmpi_master ) call system_clock(t3)
    !if ( lmpi_master ) t3 = mpi_wtime()
    if ( lmpi_master ) t3 = omp_get_wtime()
    select case(nb)
    case(5)
      call point_solve_5()
    case(:4,6:)
      write(*,*) 'That giraffed-shaped hack removed point_solve_n!'
    end select
    !if ( lmpi_master ) call system_clock(t4)
    !if ( lmpi_master ) t4 = mpi_wtime()
    if ( lmpi_master ) t4 = omp_get_wtime()
    if ( lmpi_master ) write(*,*) 'Sweep msec on master = ', (t4-t3)
  end do

  !call system_clock(t2)
  !if (lmpi_master .and. timing) t2 = mpi_wtime()
  if (lmpi_master .and. timing) t2 = omp_get_wtime()

  if (lmpi_master .and. timing) write(*,*) 'Total msec taken on master = ', (t2-t1)

  call lmpi_finalize()

end program main
